/************************************************************************
* Copyright (c) 1995 Valentin Shopov, Blagoevgrad, Bulgaria		*
*									*
* Permission to use, copy, modify, and distribute this material for any	*
* purpose and without fee is hereby granted, provided that the above	*
* copyright notice and this permission notice appear in all copies, and	*
* that the name of the author not be used in advertising or publicity 	*
* pertaining to this material without the specific, prior written 	*
* permission of the author.						*
*									*
* THE AUTHOR MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY	*  
* OF THIS MATERIAL FOR ANY PURPOSE. IT IS PROVIDED "AS IS", WITHOUT ANY	* 
* EXPRESS OR IMPLIED WARRANTIES.					*
*************************************************************************/

/**************************ciilib1.3.0***********************************
 *									*
 *        Character to Inline Image Conversion Library (CIILIB)		*
 *									*
 *	Encoder for Character by Inline Image of HTML on HTTP		*
 *			by Valentin Shopov				*
 *			valentin@rs.aubg.bg				*
 ************************************************************************/

#include <stdio.h>
#include <memory.h>
#include "cii.h"

/*////////////////////////////////////////////////////////////////////// */



/*
 *    CII output image type selector
 */
CII_setitype (sitype)
     char *sitype;
{
  ;
}

/*
 *    text to CII encoder
 */
CII_toCII (any_text, cii_text, ctype, charsets, urlbase)
     unsigned char *any_text, *cii_text;
     char *ctype, *charsets[];
     char *urlbase;
{
  static _CII_toCII ();
  return _CII_toCII (any_text, cii_text, ctype, charsets, urlbase);
}

/*
 *    CII font server
 */
CII_putHTTP (out, vno, cii_code)
     FILE *out;
     char *cii_code;
{
  return CII_putHTTP_GIF (out, vno, cii_code);
}
/*////////////////////////////////////////////////////////////////////// */

static int isTag, isTitle, isAnchor, isMozilla, isH1, TagBufLen;
static int isKOI;
static char *inHTMLcs = (char *) 0, *Hchar = "";
static unsigned char TagBuf[2048];

extern struct CharCodesXBM CII_CC[];

static int
CharsetID (abbr)
     char *abbr;
{
  int i;
  char *nick;

  for (i = 0; nick = CII_CC[i].name; i++)
    {
      if (strncasecmp (nick, abbr, strlen (abbr)) == 0)
	return i;
      if (strcasecmp (abbr, CII_CC[i].regname) == 0)
	return i;
    }
  return -1;
}


static int
to256c (csname)
     char *csname;
{
  int cc_id;

  if ((cc_id = CharsetID (csname)) == -1)
    return 0;
  return CII_CC[cc_id].is2bytes == 0;
}


static unsigned char
koi8tocyr (kchar)
     unsigned char kchar;
{
  char *cyrch = "\356\320\321\346\324\325\344\323\345\330\331\332\333\334\
\335\336\337\357\340\341\342\343\326\322\354\353\327\350\355\351\347\352";
  if (kchar >= 0xc0 && kchar <= 0xdf)
    return cyrch[kchar - 0xc0];
  else if (kchar >= 0xe0 && kchar <= 0xff)
    return cyrch[kchar - 0xe0] - 32;
  else
    return kchar;
}



static int
strncscmp (str1, str2, snum)
     char *str1, *str2;
     int snum;
{
  char ch1;
  int i, j, isSPC;
  for (i = 0, j = 0, isSPC = 0; i < snum && str1[j]; i++)
    {
      for (; str1[j] <= ' ' && str1[j];)
	{
	  j++;
	  isSPC = 1;
	}
      ch1 = str1[j];
      if (ch1 >= 'a' && ch1 <= 'z')
	ch1 -= 32;
      if (isSPC)
	{
	  isSPC = 0;
	  if (ch1 == '=' && str2[i] == '=')
	    {
	      j++;
	      continue;
	    }
	  if (str2[i - 1] == '=')
	    {
	      i--;
	      continue;
	    }
	  if (ch1 == '>' && str2[i] == '>')
	    {
	      j++;
	      continue;
	    }

	  if (str2[i] != ' ')
	    return 1;
	}
      else
	{
	  if (str2[i] != ch1)
	    return 1;
	  j++;
	}
    }
  return 0;
}

static int
isN256C (src)
     unsigned char *src;
{
  register unsigned char *sp;
  int is8bit = 0;

  for (sp = src; *sp; sp++)
    {
      if (*sp >= 0x80)
	{
	  is8bit++;
	  if (*sp < 0xa0)
	    return 1;
	}
    }
  return (is8bit) ? 0 : -1;
}

static int
rvrsBUF (src)
     unsigned char *src;
{
  register unsigned char *sp, *dp;
  unsigned char *tmpb;
  tmpb = (unsigned char *) strdup (src);
  for (sp = (unsigned char *) strrchr (tmpb, 0) - 1, dp = src; *dp; sp--, dp++)
    *dp = *sp;
  free (tmpb);
}

/*////////////////////////////////////////////////////////////////////// */
#include "gifcii.c"
#include "gifconv.c"
/*////////////////////////////////////////////////////////////////////// */

extern char *HTTP_getRequestField ();
unsigned char *
EUC_TO_CII (src, dst, ctype, charsetC, ubase)
     unsigned char *src, *dst;
     char *ctype, *charsetC, *ubase;
{
  register unsigned char *sp, *dp;
  char *charset, *charsetsII, *Achar, *ltPTR, *gtPTR;
  unsigned char t256c;
  char inHTMLCC[32];
  int isPlain, do256, AncLen, i, chW, chH;
  char UserAGENT[1024];
  char *mozi;

  HTTP_getRequestField ("User-Agent", UserAGENT, sizeof (UserAGENT));
  if (strstr (UserAGENT, "Emacs-W3") || strstr (UserAGENT, "Lynx"))
    {
      strcpy (dst, src);
      return dst;
    }

  if ((mozi = (char*)strstr (UserAGENT, "Mozilla")) && strcmp (mozi, "Mozilla/1.1") >= 0)
    isMozilla = 1;
  else
    isMozilla = 0;

  GIFCII_init (isMozilla);

  isPlain = strcasecmp (ctype, "text/plain") == 0;

  charsetsII = (inHTMLcs) ? inHTMLcs : charsetC;

  charset = (isKOI == 1) ? CII_CC[CharsetID ("cyr")].regname : CII_CC[CharsetID (charsetsII)].regname;
  if (charset == NULL)
    {
      strcpy (dst, src);
      return dst;
    }

  do256 = to256c (charset);

  ltPTR = (char *) strrchr (src, '<');
  gtPTR = (char *) strrchr (src, '>');

  if ((gtPTR && ltPTR - gtPTR > 0) || ((TagBufLen || ltPTR) && !gtPTR))
    {
      for (i = 0; src[i]; i++)
	{
	  TagBuf[TagBufLen++] = src[i];
	}
      TagBuf[TagBufLen] = '\0';
      *dst = '\0';
      return dst;
    }

  if (TagBufLen)
    {
      for (i = 0; src[i]; i++)
	{
	  TagBuf[TagBufLen++] = src[i];
	}
      TagBuf[TagBufLen] = '\0';
    }



  for (sp = (TagBufLen) ? TagBuf : src, dp = dst; *sp;)
    {
      if (*sp == '<' && !isTag && !isPlain)
	{

	  isTag = 1;
	  if (!isH1 && !strncscmp (sp, "<H1>", 4))
	    {
	      isH1 = 1;
	      Hchar = "h";
	    }
	  else if (GIFCII_switch_H (sp))
	    ;
	  else if (GIFCII_switch_Color (sp, isMozilla))
	    ;
	  else if (isH1 && !strncscmp (sp, "</H1>", 5))
	    {
	      isH1 = 0;
	      Hchar = "";
	    }
	  else if (!isAnchor && !strncscmp (sp, "<A HREF=", 8))
	    {
	      isAnchor = 1;
	    }
	  else if (isAnchor && !strncscmp (sp, "</A>", 4))
	    {
	      isAnchor = 0;
	      isTag = 0;
	    }
	  else if (!strncasecmp (sp, "<!CIICC=", 8))
	    {
	      sscanf (sp + 8, "%[^>]", inHTMLCC);
	      free ((void *) inHTMLcs);
	      if (!strncmp (inHTMLCC, "koi8", 4) && isKOI != 2)
		{
		  inHTMLcs = (char *) strdup ("koi8");
		  isKOI = 1;
		}
	      else
		inHTMLcs = (char *) strdup (CII_CC[CharsetID (inHTMLCC)].regname);

	      charsetsII = (inHTMLcs) ? inHTMLcs : charsetC;

	      if (!strncmp (charsetsII, "koi8", 4) && isKOI != 2)
		{
		  charset = CII_CC[CharsetID ("cyr")].regname;
		  do256 = to256c ("cyr");
		  isKOI = 1;
		}
	      else
		{
		  isKOI = (isKOI == 2) ? isKOI : 0;
		  charset = CII_CC[CharsetID (charsetsII)].regname;
		  do256 = to256c (charsetsII);
		}
	    }
	  else if (!strncscmp (sp, "<!/CIICC>", 9))
	    {
	      free ((void *) inHTMLcs);
	      inHTMLcs = (char *) 0;
	      if (!strncmp (charsetC, "koi8", 4) && isKOI != 2)
		{
		  charset = CII_CC[CharsetID ("cyr")].regname;
		  do256 = to256c ("cyr");
		  isKOI = 1;
		}
	      else
		{
		  isKOI = (isKOI == 2) ? isKOI : 0;
		  charset = CII_CC[CharsetID (charsetC)].regname;
		  do256 = to256c (charsetC);
		}

	    }
	  else if (!isTitle && !strncscmp (sp, "<TITLE>", 7))
	    isTitle = 1;
	  else if (isTitle && !strncscmp (sp, "</TITLE>", 8))
	    isTitle = 0;
	}
      if (isTag || isTitle)
	{
	  if (*sp == '>' && isTag)
	    isTag = 0;
	  *dp++ = *sp++;
	}
      else
	{
	  if (*sp & 0x80)
	    {
	      Achar = "";

	      GIFCII_setfsize (&chH, &chW, &Achar, charset);


	      if (((do256) ? 1 : (sp[1] & 0x80)))
		{
		  sp = GIFCII_putCII (dp, sp, chH, chW, ubase, charset, Achar, isMozilla, do256);
		  dp += strlen (dp);
		}
	      else
		sp++;

	    }
	  else
	    {
	      if (isPlain)
		{
		  switch (*sp)
		    {
		    case '<':
		      sprintf (dp, "&lt;");
		      dp += 4;
		      sp++;
		      break;
		    case '>':
		      sprintf (dp, "&gt;");
		      dp += 4;
		      sp++;
		      break;
		    case '&':
		      sprintf (dp, "&amp;");
		      dp += 5;
		      sp++;
		      break;
		    case '"':
		      sprintf (dp, "&quot;");
		      dp += 6;
		      sp++;
		      break;
		    default:
		      *dp++ = *sp++;
		      break;
		    }
		}
	      else
		{
		  *dp++ = *sp++;
		}
	    }
	}
    }
  *dp = 0;
  if (TagBufLen)
    TagBufLen = 0;

  return dst;
}

#define ESC 	'\033'
#define TO1B 	','
#define TO2B	'$'
#define TOASC1	'('
#define TOASC2	'B'
#define TOJAP	'B'
#define TOCHI	'A'
#define TOCYR 	'L'
#define TOGRE 	'F'
#define TOHEB 	'H'
#define TOKOR1	'('
#define TOKOR2  'C'
#define R2L1 	'['
#define R2L3	']'
#define R2LIN	'2'
#define R2LOUT 	'0'

ISO2022_toCII (src, dst, ctype, charsets, ubase)
     unsigned char *src, *dst;
     char *ctype, *charsets[], *ubase;
{
  unsigned char *tmpbuf, *tmpdst, *euc;
  register unsigned char *sp, *dp;
  static char *Charsets[] =
  {"jp", "cyrillic", "greek", "hebrew", "chinese", "korean", 0};
  int csn = 0, csnprev, isR2L = 0, isAscii = 1;

  tmpdst = dst;
  *tmpdst = '\0';
  tmpbuf = (unsigned char *) malloc (strlen (src) + 1);
  for (sp = src, dp = tmpbuf; *sp;)
    {
      if (*sp == ESC)
	{
	  csnprev = csn;
	  if (sp[1] == TO1B)
	    {
	      switch (sp[2])
		{
		case TOCYR:
		  csn = 1;
		  break;
		case TOGRE:
		  csn = 2;
		  break;
		case TOHEB:
		  csn = 3;
		  break;
		default:
		  csn = 0;
		}
	      if (csn)
		{
		  isAscii = 0;
		  sp += 3;
		}
	      else
		{
		  *dp++ = *sp++;
		  *dp = '\0';
		}
	      if (csn && csn != csnprev)
		{
		  *dp = '\0';
		  if (!csnprev)
		    {
		      euc = (unsigned char *) strdup (tmpbuf);
		      TO_EUC (euc, tmpbuf);
		      free (euc);
		    }
		  if (isR2L)
		    rvrsBUF (tmpbuf);
		  EUC_TO_CII (tmpbuf, tmpdst, ctype, Charsets[csnprev], ubase);
		  tmpdst += strlen (tmpdst);
		  dp = tmpbuf;
		}


	    }

	  else if (sp[1] == TO2B)
	    {
	      if (sp[2] == TOCHI)
		{
		  csn = 4;
		  sp += 3;
		  isAscii = 0;
		}
	      else if (sp[2] == TOKOR1 && sp[3] == TOKOR2)
		{
		  csn = 5;
		  sp += 4;
		  isAscii = 0;
		}
	      else
		{
		  csn = 0;
		  *dp++ = *sp++;
		  *dp = '\0';
		}

	      if (csn && csn != csnprev)
		{
		  *dp = '\0';
		  if (!csnprev)
		    {
		      euc = (unsigned char *) strdup (tmpbuf);
		      TO_EUC (euc, tmpbuf);
		      free (euc);
		    }
		  if (isR2L)
		    rvrsBUF (tmpbuf);
		  EUC_TO_CII (tmpbuf, tmpdst, ctype, Charsets[csnprev], ubase);
		  tmpdst += strlen (tmpdst);
		  dp = tmpbuf;
		}


	    }

	  else if (csn && sp[1] == TOASC1 && sp[2] == TOASC2)
	    {
	      sp += 3;
	      isAscii = 1;
	    }
	  else if (sp[1] == R2L1 && sp[3] == R2L3)
	    {
	      if (sp[2] == R2LIN)
		{
		  sp += 4;
		  if (!isR2L)
		    {
		      isR2L = 1;
		      *dp = '\0';
		      EUC_TO_CII (tmpbuf, tmpdst, ctype, Charsets[csn], ubase);
		      tmpdst += strlen (tmpdst);
		      dp = tmpbuf;
		    }
		}
	      else if (sp[2] == R2LOUT)
		{
		  sp += 4;
		  if (isR2L)
		    {
		      *dp = '\0';
		      if (isR2L)
			rvrsBUF (tmpbuf);
		      EUC_TO_CII (tmpbuf, tmpdst, ctype, Charsets[csn], ubase);
		      tmpdst += strlen (tmpdst);
		      dp = tmpbuf;
		      isR2L = 0;
		    }
		}
	      else
		{
		  *dp++ = *sp++;
		  *dp = '\0';
		}

	    }
	  else
	    {
	      *dp++ = *sp++;
	      *dp = '\0';
	    }

	}
      else
	{
	  *dp = (csn && !isAscii && *sp > 0x20 && *sp <= 0x7f) ? *sp | 0x80 : *sp;
	  dp++;
	  *dp = '\0';
	  sp++;
	}
    }

  if (!csn)
    {
      euc = (unsigned char *) strdup (tmpbuf);
      TO_EUC (euc, tmpbuf);
      free (euc);
    }
  EUC_TO_CII (tmpbuf, tmpdst, ctype, Charsets[csn], ubase);
  free (tmpbuf);
}

static
_CII_toCII (any, gr, ctype, charsets, ubase)
     unsigned char *any, *gr;
     char *ctype, *charsets[];
     char *ubase;
{
  unsigned char *euc;
  int i, strC;
  char *charsetsII;

  strC = isN256C (any);
  charsetsII = (inHTMLcs) ? inHTMLcs : charsets[0];
  if (CharsetID ("koi8") != -1)
    isKOI = 2;
  else if (!strncasecmp (charsetsII, "koi8", 4))
    isKOI = 1;
  else
    isKOI = 0;
  if ((strC >= 0 && isKOI) || (!strC && to256c (charsetsII)))
    EUC_TO_CII (any, gr, ctype, charsets[0], ubase);
  else
    {
      if (strC < 0 && strchr (any, 0x1b))
	ISO2022_toCII (any, gr, ctype, charsets, ubase);

      else
	{
	  if (CharsetID (charsetsII) == CharsetID ("jp"))
	    {
	      euc = (unsigned char *) malloc (strlen (any) + 1);
	      TO_EUC (any, euc);
	      EUC_TO_CII (euc, gr, ctype, charsets[0], ubase);
	      free (euc);
	    }
	  else
	    EUC_TO_CII (any, gr, ctype, charsets[0], ubase);

	}
    }
}
